@@ -146,7 +146,7 @@ class LinearAttention(nn.Module):
         """
         query = self.activation(query)
         numerator = torch.matmul(query, matrix)
-        denominator = torch.matmul(query, normalizer.unsqueeze(-1)) + 1e-9
+        denominator = torch.matmul(query, normalizer.unsqueeze(-1)) + 1e-5
         return numerator / denominator
 
     def make_kernel(self,
@@ -343,17 +343,32 @@ class WriteMemory(nn.Module):
                  num_write_heads: int,
                  num_memories: int,
                  dropout_rate: float,
+                 max_write_factor: float,
                  linear_kernel_activation: Callable[[torch.Tensor], torch.Tensor],
                  dtype: torch.dtype,
                  device: torch.device
-
                  ):
+        """
+        :param d_model: The model dim
+        :param d_address: The address di
+        :param d_memory: the memory dim
+        :param num_write_heads: num write heads
+        :param num_memories: num memores
+        :param dropout_rate: write selection logit dropout
+        :param max_write_factor: Maximum rate we can write to a memory with
+        - Note: While max is 1.0, for numeric stability you should keep it lower like 0.7
+        :param linear_kernel_activation: The kernel activation
+        :param dtype: The dtype
+        :param device: The device.
+        """
+
         super().__init__()
         self.d_model = d_model
         self.d_address = d_address
         self.d_memory = d_memory
         self.num_write_heads = num_write_heads
         self.linear_kernel_activation = linear_kernel_activation
+        self.max_write_factor = max_write_factor
         self.__metainfo = DeviceDtypeWatch(device=device, dtype=dtype)
 
         # Create the linear attn mechanism.
@@ -392,7 +407,7 @@ class WriteMemory(nn.Module):
         # Create the probability projectors and interpolation logits
         self.interpolation_logits = initialize_based_on_half_lives([num_memories, d_address],
                                                                    1,
-                                                                   512)
+                                                                   60)
         self.write_logits = nn.Linear(d_address, 1, dtype=dtype, device=device)
         self.dropout_logits = DropoutLogits(dropout_rate)
 
@@ -463,6 +478,10 @@ class WriteMemory(nn.Module):
         interpolation_factor = torch.sigmoid(self.interpolation_logits)  # (num_memories, d_address)
         write_factor = write_probability * interpolation_factor
 
+        # In order to prevent numeric explosion, we must limit the maximum available write factor
+        # Otherwise, the reverse pass can divide by zero
+        write_factor = self.max_write_factor*write_factor
+
         # Create kernel update. Unsqueeze to fit linear attn mechanism format.
 
         update = self.linear_attn.make_kernel(key.unsqueeze(-2), value.unsqueeze(-2))
@@ -490,8 +509,8 @@ class WriteMemory(nn.Module):
         matrix_update, normalizer_update = update
 
         # Run the inverse of the original update factor.
-        normalizer = (next_normalizer - normalizer_update * write_factor) / (1 - write_factor)
-        matrix = (next_matrix - matrix_update * write_factor.unsqueeze(-1)) / (1 - write_factor.unsqueeze(-1))
+        normalizer = (next_normalizer  - normalizer_update * write_factor)/(1 - write_factor)
+        matrix = (next_matrix  - matrix_update * write_factor.unsqueeze(-1))/(1 - write_factor.unsqueeze(-1))
         cum_prob = next_cum_prob - write_factor.mean(dim=-1)
 
         # Mask out anything that was not able to be updated
@@ -524,7 +543,7 @@ class WriteMemory(nn.Module):
 
         # Get the key common information
         normalizer = last_normalizer * (1 - write_factor) + normalizer_update * write_factor
-        matrix = last_matrix * (1 - write_factor.unsqueeze(-1)) + matrix_update * write_factor.unsqueeze(-1)
+        matrix = last_matrix * (1-write_factor.unsqueeze(-1)) + matrix_update * write_factor.unsqueeze(-1)
         cum_prob = last_cum_prob + write_factor.mean(dim=-1)
 
         # Mask out anything that was not able to be updated
@@ -569,6 +588,7 @@ class FastLinearMemory(nn.Module):
                  num_memories: int,
                  dropout_rate: int,
                  linear_kernel_activation: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
+                 max_write_factor: float = None,
                  dtype: Optional[torch.dtype] = None,
                  device: Optional[torch.device] = None,
                  ):
@@ -580,13 +600,20 @@ class FastLinearMemory(nn.Module):
         :param num_write_heads: The number of write heads. 4-8 is fine.
         :param num_memories: The number of memories under consideration. Should be large
         :param linear_kernel_activation: The activation kernel for linear attention. Defaults to elu
+        :param max_write_factor: The maximum probability, between 0 and 1, that can be committed at once
+                                 to a memory position
+                                 - It is required to be less than 1.0 for numeric reasons.
+                                 - Default is 0.6
+                                 - If numeric divergence is an issue, try lowering it a bit.
         :param dtype: defaults to float32
         :param device: defaults to cpu
         """
 
         # Defaults
         if linear_kernel_activation is None:
-            linear_kernel_activation = F.elu
+            linear_kernel_activation = F.softplus
+        if max_write_factor is None:
+            max_write_factor = 0.6
 
         super().__init__()
 
@@ -599,7 +626,8 @@ class FastLinearMemory(nn.Module):
         self.memory_reader = ReadMemory(d_model, d_address, d_memory, num_read_heads, linear_kernel_activation,
                                         device=device, dtype=dtype)
         self.memory_writer = WriteMemory(d_model, d_address, d_memory, num_write_heads, num_memories,
-                                         dropout_rate, linear_kernel_activation, dtype=dtype, device=device)
+                                         dropout_rate, max_write_factor, linear_kernel_activation,
+                                         dtype=dtype, device=device)
 
         # Create the device metainfo watch.
         self.__metainfo = DeviceDtypeWatch(device=device, dtype=dtype)
@@ -651,18 +679,15 @@ class FastLinearMemory(nn.Module):
             original_memory = self.memory_writer.reverse_memory(update, write_factor, batch_mask, next_memory)
 
         def setup_grads(tensor: torch.Tensor)->torch.Tensor:
-            if torch.is_grad_enabled():
-                tensor.requires_grad_(True)
-                tensor = tensor.clone()
-                tensor.retain_grad()
+            tensor = nn.Parameter(tensor, requires_grad=True)
+            tensor.retain_grad()
             return tensor
+
         original_memory = parallel_pytree_map(setup_grads, original_memory)
 
-        # Manually compute the write, then the read.
+        # Manually complete the read
         next_memory = self.memory_writer.advance_memory(update, write_factor, batch_mask, original_memory)
         read = self.memory_reader(tensor, self.addresses, next_memory)
-
-        # Return the results
         return (read, next_memory), original_memory
     def forward(self,
                 tensor: torch.Tensor,
@@ -681,6 +706,8 @@ class FastLinearMemory(nn.Module):
         - The response tensor. Shape (..., d_model)
         - The new memory state
         """
+
+
         update, write_factor = self.memory_writer.compute_common(tensor, tensor, self.addresses)
         memory = self.memory_writer.advance_memory(update, write_factor, batch_mask, memory)
         read = self.memory_reader(tensor, self.addresses, memory)
